"""
    # Workaround / say if using cached Charm present
    # Write out resolved deployment file with includes & overrides
    # Ensure bootstrap

    # Deployment validate
"""

import logging
import optparse
import os

import sys
import time

from juju_deployer import utils

from juju_deployer.config import ConfigStack
from juju_deployer.environment import Environment
from juju_deployer.importer import Importer
from juju_deployer.exception import ErrorExit


def setup_parser():
    parser = optparse.OptionParser()
    parser.add_option('-c', '--config',
                      help=('File containing deployment(s) json config. This '
                            'option can be repeated, with later files '
                            'overriding values in earlier ones.'),
                      dest='configs', action='append')
    parser.add_option('-d', '--debug', help='Enable debugging to stdout',
                      dest="debug",
                      action="store_true", default=False)
    parser.add_option('-L', '--local-mods',
                      help='Disallow deployment of locally-modified charms',
                      dest="no_local_mods", default=True, action='store_false')
    parser.add_option('-u', '--update-charms',
                      help='Update existing charm branches',
                      dest="update_charms", default=False, action="store_true")
    parser.add_option('-l', '--ls', help='List available deployments',
                      dest="list_deploys", action="store_true", default=False)
    parser.add_option('-D', '--destroy-services',
                      help='Destroy all services (do not terminate machines)',
                      dest="destroy_services", action="store_true",
                      default=False)
    parser.add_option('-T', '--terminate-machines',
                      help=('Terminate all machines but the bootstrap node.  '
                            'Destroy any services that exist on each'),
                      dest="terminate_machines", action="store_true",
                      default=False)
    parser.add_option('-t', '--timeout',
                      help=('Timeout (sec) for entire deployment '
                            '(45min default)'),
                      dest='timeout', action='store', type='int', default=2700)
    parser.add_option("-f", '--find-service', action="store", type="string",
                      help=('Find hostname from first unit of a specific '
                            'service.'),
                      dest="find_service")
    parser.add_option("-b", '--branch-only', action="store_true",
                      help='Update vcs branches and exit.',
                      dest="branch_only")
    parser.add_option('-s', '--deploy-delay', action='store', type='float',
                      help=("Time in seconds to sleep between 'deploy' "
                            "commands to allow machine provider to process "
                            "requests. This delay is also enforced between "
                            "calls to terminate_machine"),
                      dest="deploy_delay", default=0)
    parser.add_option('-e', '--environment', action='store', dest='juju_env',
                      help='Deploy to a specific Juju environment.',
                      default=os.getenv('JUJU_ENV'))
    parser.add_option('-o', '--override', action='append', type='string',
                      help=('Override *all* config options of the same name '
                            'across all services.  Input as key=value.'),
                      dest='overrides', default=None)
    parser.add_option('-v', '--verbose', action='store_true', default=False,
                      dest="verbose", help='Verbose output')
    parser.add_option('-W', '--watch', action="store_true", default=False,
                      help='Watch environment changes on console.',
                      dest="watch"),
    parser.add_option('-r', "--retry", default=0, type=int, dest="retry_count",
                      help=("Resolve unit errors via retry. Either standalone "
                            "or in a deployment"))
    parser.add_option('-w', '--relation-wait', action='store', dest='rel_wait',
                      default=60, type=int,
                      help=('Number of seconds to wait before checking for '
                            'relation errors after all relations have been '
                            'added  and subordinates started. (default: 60)'))

    return parser


def main():
    stime = time.time()
    try:
        run()
    except ErrorExit:
        logging.getLogger('deployer.cli').info(
            "Deployment stopped. run time: %s", time.time() - stime)
        sys.exit(1)


def run():
    parser = setup_parser()
    (options, args) = parser.parse_args()

    # Debug implies watching and verbose
    if options.debug and not options.watch:
        options.watch = options.verbose = True
    utils.setup_logging(options.verbose, options.debug)

    log = logging.getLogger("deployer.cli")
    start_time = time.time()

    env = Environment(options.juju_env)
    config = ConfigStack(options.configs or [])

    # Destroy services and exit
    if options.destroy_services or options.terminate_machines:
        log.info("Resetting environment...")
        env.connect()
        env.reset(terminate_machines=options.terminate_machines,
                  terminate_delay=options.deploy_delay,
                  watch=options.watch)
        log.info("Environment reset in %0.2f", time.time() - start_time)
        sys.exit(0)

    # Display service info and exit
    if options.find_service:
        address = env.get_service_address(options.find_service)
        if address is None:
            log.error("Service not found %r", options.find_service)
            sys.exit(1)
        elif not address:
            log.warning("Service: %s has no address for first unit",
                        options.find_service)
        else:
            log.info("Service: %s address: %s", options.find_service, address)
        sys.exit(0)

    # Just resolve/retry hooks in the environment
    if not args and options.retry_count:
        log.info("Retrying hooks for error resolution")
        env.connect()
        env.resolve_errors(
            options.retry_count, watch=options.watch, timeout=options.timeout)

    # Arg check on config files and deployment name.
    if not options.configs:
        log.error("Config files must be specified")
        sys.exit(1)

    config.load()

    # Just list the available deployments
    if options.list_deploys:
        log.info("%s", " ".join(sorted(config.keys())))
        sys.exit(0)

    # Do something to a deployment
    if not args:
        log.error(
            "Deployment name must be specified. available: %s", tuple(
                sorted(config.keys())))
        sys.exit(1)

    deploy_name = args[0]
    deployment = config.get(deploy_name)

    # Do this thing
    log.debug("Starting deployment of %s", deploy_name)
    Importer(env, deployment, options).run()

    # Deploy complete
    log.info("Deployment complete in %0.2f seconds" % (
        time.time() - start_time))


if __name__ == '__main__':
    main()
