import errno
import logging
import os
import socket
import time

from os.path import join as path_join

from juju_deployer import utils
from juju_deployer import client
from juju_deployer.exception import ErrorExit

from jujuclient import Environment as EnvironmentClient, UnitErrors


class Endpoint(object):

    def __init__(self, ep):
        self.ep = ep
        self.name = None
        if ":" in self.ep:
            self.service, self.name = self.ep.split(":")
        else:
            self.service = ep


def get_juju_impl(name):
    return client.CLIClient(name)


class Environment(object):

    log = logging.getLogger("deployer.env")

    def __init__(self, name):
        # IMPL
        self.name = name
        self.client = get_juju_impl(name)

    def close(self):
        # IMPL
        if self.client:
            self.client.close()

    def connect(self):
        self.client.connect()
        self.log.debug("Connected to environment")
        return
        # IMPL
        if not self.api_endpoint:
            # should really have a cheaper/faster way of getting the endpoint
            # ala juju status 0 for a get_endpoint method.
            self.get_cli_status()
        while True:
            try:
                self.client = EnvironmentClient(self.api_endpoint)
            except socket.error, e:
                if e.errno != errno.ETIMEDOUT:
                    raise
                continue
            else:
                break
        self.client.login(self._get_token())
        self.log.debug("Connected to environment")

    def _get_token(self):
        # IMPL
        with open(self.env_config_path) as fh:
            config = utils.yaml_load(fh.read())
            if self.name:
                token = config['environments'][self.name]['admin-secret']
            else:
                env_name = config.get('default')
                if env_name is None:
                    if len(config['environments'].keys()) == 1:
                        env_name = config['environments'].keys().pop()
                    else:
                        self.log.error("Ambigious operation environment")
                        raise ErrorExit()
                token = config['environments'][env_name]['admin-secret']
            return token

    @property
    def env_config_path(self):
        # COMMON
        jhome = os.environ.get("JUJU_HOME")
        if jhome is None:
            jhome = path_join(os.environ.get('HOME'), '.juju')
        env_config_path = path_join(jhome, 'environments.yaml')
        return env_config_path

    def _named_env(self, params):
        # IMPL?
        if self.name:
            params.extend(["-e", self.name])
        return params

    def deploy(self, name, charm_url, repo=None, config=None,
               constraints=None, num_units=1, force_machine=None):
        self.client.deploy(name, charm_url, repo, config, constraints,
                           num_units=1, force_machine=None)
        return
        # COMMON
        params = self._named_env(["juju", "deploy"])

        with utils.temp_file() as fh:
            if config:
                fh.write(utils.yaml_dump({name: config}))
                fh.flush()
                params.extend(["--config", fh.name])
            if constraints:
                params.extend(['--constraints', constraints])
            if num_units != 1:
                params.extend(["--num-units", str(num_units)])
            if charm_url.startswith('local'):
                if repo == "":
                    repo = "."
                params.extend(["--repository=%s" % repo])
            if force_machine is not None:
                params.extend["--force-machine=%d" % force_machine]

            params.extend([charm_url, name])
            utils.check_call(
                params, '', self.log, "Error deploying service %r", name)

    def add_units(self, service_name, num_units):
        return self.client.add_units(service_name, num_units)

    def add_relation(self, endpoint_a, endpoint_b):
        return self.client.add_relation(endpoint_a, endpoint_b)

    def terminate_machine(self, mid, wait=False):
        return self.client.terminate_machine(mid, wait)

        if int(mid) == 0:
            raise RuntimeError("Can't terminate machine 0")
        params = self._named_env(["juju", "terminate-machine"])
        params.append(mid)
        utils.check_call(params, "", self.log,
                         "Error terminating machine %r" % mid)

    def get_service_address(self, svc_name):
        status = self.get_cli_status()
        if svc_name not in status['services']:
            self.log.warning("Service %s does not exist", svc_name)
            return None
        units = status['services'][svc_name].get('units', {})
        unit_keys = list(sorted(units.keys()))
        if unit_keys:
            return units[unit_keys[0]].get('public-address', '')
        self.log.warning("Service %s has no units")

    def get_cli_status(self):
        return self.client.get_cli_status()
        # IMPL
        params = self._named_env(["juju", "status"])
        output = utils.check_call(
            params, "", self.log, "Error getting status, is it bootstrapped?")
        status = utils.yaml_load(output)
        # Opportunistic, see connect method comment.
        if not self.api_endpoint:
            self.api_endpoint = "wss://%s:17070/" % (
                status["machines"]["0"]["dns-name"])
        return status

    def reset(self, terminate_machines=False, terminate_delay=0,
              timeout=360, watch=False):
        # IMPL
        """Destroy/reset the environment."""
        status = self.client.status()
        destroyed = False
        for s in status.get('services', {}).keys():
            self.log.debug(" Destroying service %s", s)
            self.client.destroy_service(s)
            destroyed = True

        if destroyed:
            # Mark any errors as resolved so destruction can proceed.
            self.resolve_errors()

            # Wait for units
            self.wait_for_units(timeout, "removed", watch=watch)

        # The only value to not terminating is keeping the data on the machines
        # around.
        if not terminate_machines:
            self.log.info(" *juju-core machines are not reusable for units")
            return

        # Terminate machines
        self.log.debug(" Terminating machines")
        for mid in status['machines'].keys():
            if mid == "0":
                continue
            self.log.debug("  Terminating machine %s", mid)
            self.client.terminate_machine(mid)
            if terminate_delay:
                self.log.debug("  Waiting for terminate delay")
                time.sleep(terminate_delay)

    def resolve_errors(self, retry_count=0, timeout=600, watch=False, delay=5):
        """Resolve any unit errors in the environment.

        If retry_count is given then the hook execution is reattempted. The
        system will do up to retry_count passes through the system resolving
        errors.

        If retry count is not given, the error is marked resolved permanently.
        """
        # IMPL
        etime = time.time() + timeout
        count = resolved = 0
        while True:
            status = self.client.status()
            for s in status.get('services', {}).keys():
                for uid, u in status['services'][s].get('units', {}).items():
                    if u['agent-state'] == 'error':
                        self.log.info("  Retrying unit error %s" % (uid))
                        self.client.resolved(uid, retry=bool(retry_count))
                        resolved = True
                    for uid, u in u.get('subordinates', {}).items():
                        if u['agent-state'] == 'error':
                            self.log.info("  Retrying subordinate error %s" %
                                          (uid))
                            self.client.resolved(uid, retry=bool(retry_count))
                            resolved = True
            if not retry_count or not resolved:
                self.log.debug("  No errors found.")
                return

            w_timeout = etime - time.time()
            if w_timeout < 0:
                self.log.error("Timeout reached while resolving errors")
                raise ErrorExit()

            if retry_count:
                time.sleep(delay)

            count += 1
            try:
                self.wait_for_units(
                    timeout=int(w_timeout), watch=True, no_exit=True)
            except UnitErrors, e:
                if retry_count == count:
                    self.log.info(
                        " Retry count %d exhausted, but units still in error "
                        "(%s)",
                        retry_count, " ".join(u['Name'] for u in e.errors))
                    return
            else:
                return

    def status(self):
        return self.client.status()

    def wait_for_units(
            self, timeout, goal_state="started", watch=False, no_exit=False):
        """Wait for units to reach a given condition.
        """
        return self.client.wait_for_units(timeout, goal_state, watch, no_exit)
        # IMPL
        callback = watch and self._delta_event_log or None
        self.log.debug("Waiting for units to be %s", goal_state)
        try:
            self.client.wait_for_units(timeout, goal_state, callback=callback)
        except UnitErrors, e:
            error_units = [
                "unit: %s: machine: %s agent-state: %s details: %s" % (
                    u['Name'], u['MachineId'], u['Status'], u['StatusInfo']
                )
                for u in e.errors]
            if no_exit:
                raise
            self.log.error("The following units had errors:\n   %s" % (
                "   \n".join(error_units)))
            raise ErrorExit()

    def _delta_event_log(self, et, ct, d):
        # event type, change type, data
        name = d.get('Name', d.get('Id', 'unknown'))
        state = d.get('Status', d.get('Life', 'unknown'))
        if et == "relation":
            name = self._format_endpoints(d['Endpoints'])
            state = "created"
            if ct == "remove":
                state = "removed"
        self.log.debug(
            " Delta %s: %s %s:%s", et, name, ct, state)

    def _format_endpoints(self, eps):
        if len(eps) == 1:
            ep = eps.pop()
            return "[%s:%s:%s]" % (
                ep['ServiceName'],
                ep['Relation']['Name'],
                ep['Relation']['Role'])

        return "[%s:%s <-> %s:%s]" % (
            eps[0]['ServiceName'],
            eps[0]['Relation']['Name'],
            eps[1]['ServiceName'],
            eps[1]['Relation']['Name'])
